(function() {
  var Query, Select, copy, dialects, fluid, normalize, passthrough_resolver, sys, unknown;
  var __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; }, __hasProp = Object.prototype.hasOwnProperty, __extends = function(child, parent) {
    for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; }
    function ctor() { this.constructor = child; }
    ctor.prototype = parent.prototype;
    child.prototype = new ctor;
    child.__super__ = parent.prototype;
    return child;
  }, __slice = Array.prototype.slice;
  fluid = require('./fluid');
  sys = require('sys');
  unknown = function(type, val) {
    throw new Error("Unknown " + type + ": " + val);
  };
  copy = function(thing) {
    var c, k, obj, v;
    if (thing.constructor === Array) {
      c = [];
      c.push.apply(c, thing);
      return c;
    } else if ('object' === typeof thing) {
      obj = {};
      for (k in thing) {
        v = thing[k];
        obj[k] = copy(v);
      }
      return obj;
    } else {
      return thing;
    }
  };
  dialects = exports.dialects = {
    "default": 'mysql',
    mysql: require('./mysql')
  };
  passthrough_resolver = {
    table: function(n) {
      return n;
    },
    field: function(n) {
      return n;
    }
  };
  exports.Query = Query = (function() {
    function Query(opts) {
      this.dialect = opts.dialect || dialects["default"];
      this.resolve = opts.resolver || passthrough_resolver;
      if ('string' === typeof this.dialect) {
        this.dialect = dialects[this.dialect];
      }
      this.s = {
        includedAliases: {},
        fields: {},
        tableStack: [],
        where: [],
        order: [],
        groupings: [],
        parameters: [],
        raw_where: []
      };
    }
    Query.prototype.clone = function() {
      var child, firstTable;
      firstTable = this.s.tableStack[0][0];
      child = new this.constructor(firstTable);
      child.dialect = this.dialect;
      child.resolve = this.resolve;
      child.s = copy(this.s);
      return child;
    };
    Query.prototype.aliasPair = function(table) {
      var a, t;
      if ('object' === typeof table && Object.keys(table).length === 1) {
        return ((function() {
          var _results;
          _results = [];
          for (a in table) {
            t = table[a];
            _results.push([t, a]);
          }
          return _results;
        })())[0];
      } else {
        return [table, table];
      }
    };
    Query.prototype.pushTable = function(table, alias, type, clause) {
      var _base, _ref;
      if (table === "t1t1") {
        throw new Error("here");
      }
      if (type !== 'NOP') {
        this.s.includedAliases[alias] = table;
        if ((_ref = (_base = this.s.fields)[alias]) == null) {
          _base[alias] = [];
        }
      }
      return this.s.tableStack.push([table, alias, type, clause]);
    };
    Query.prototype.lastTable = function() {
      return this.s.tableStack[this.s.tableStack.length - 1][1];
    };
    Query.prototype.includesAlias = function(a) {
      return this.s.includedAliases[a];
    };
    Query.prototype.pushParams = function(clauses) {
      var clause, _i, _len, _ref, _results;
      _results = [];
      for (_i = 0, _len = clauses.length; _i < _len; _i++) {
        clause = clauses[_i];
        _results.push(clause.op === 'multi' ? (sys.puts("pushParam recursing" + clause.clauses), this.pushParams(clause.clauses)) : clause.op === 'IN' ? (_ref = this.s.parameters).push.apply(_ref, clause.value) : this.s.parameters.push(clause.value));
      }
      return _results;
    };
    Query.prototype.visit = fluid(function(fn) {
      if (fn != null) {
        return fn.call(this, this);
      }
    });
    Query.prototype.toSql = function() {
      return this.dialect["render" + this.s.queryType](this.s);
    };
    Query.prototype.toString = function() {
      return "[Query \"" + (this.toSql().substring(0, 20)) + "\"]";
    };
    Query.prototype.execute = function(conn, cb) {
      if (conn['acquire'] != null) {
        return conn.acquire(__bind(function(c) {
          return this.execute(c, cb);
        }, this));
      } else {
        return conn.query(this.toSql(), this.s.parameters, cb);
      }
    };
    return Query;
  })();
  exports.Select = Select = (function() {
    __extends(Select, Query);
    function Select(table, opts) {
      var alias, _ref;
      if (opts == null) {
        opts = {};
      }
      Select.__super__.constructor.call(this, opts);
      this.s.queryType = 'Select';
      _ref = this.aliasPair(table), table = _ref[0], alias = _ref[1];
      this.pushTable(table, alias);
    }
    Select.prototype.from = fluid(function(alias) {
      var table;
      if (table = this.includesAlias(alias)) {
        return this.pushTable(table, alias, 'NOP');
      } else {
        return unknown('table', table);
      }
    });
    Select.prototype.fields = fluid(function() {
      var alias, f, fields, _i, _len, _results;
      fields = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
      alias = this.lastTable();
      _results = [];
      for (_i = 0, _len = fields.length; _i < _len; _i++) {
        f = fields[_i];
        _results.push(this.s.fields[alias].push(f));
      }
      return _results;
    });
    Select.prototype.field = Select.fields;
    Select.prototype.join = fluid(function(tbl, opts) {
      var alias, clause, table, type, _ref;
      if (opts == null) {
        opts = {};
      }
      _ref = this.aliasPair(tbl), table = _ref[0], alias = _ref[1];
      if (this.includesAlias(alias)) {
        throw new Error("Table alias is not unique: " + alias);
      }
      type = this.dialect.joinType(opts.type);
      clause = opts.on;
      if (clause != null) {
        if (clause.constructor !== Array) {
          clause = [clause];
        }
        clause = normalize.clauses(clause, alias, this.dialect.joinOp);
        if (clause.length > 1) {
          clause = {
            op: 'multi',
            glue: ' AND ',
            clauses: clause
          };
        } else {
          clause = clause[0];
        }
      }
      return this.pushTable(table, alias, type, clause);
    });
    Select.prototype.where = fluid(function(tbl, clause) {
      var normalized, _ref;
      if (!(clause != null)) {
        clause = tbl;
        tbl = this.lastTable();
      }
      if (this.includesAlias(tbl) == null) {
        unknown('table', tbl);
      }
      normalized = normalize.clauses([clause], tbl, this.dialect.whereOp);
      (_ref = this.s.where).push.apply(_ref, normalized);
      return this.pushParams(normalized);
    });
    Select.prototype.or = fluid(function() {
      var args, clauses;
      args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
      clauses = normalize.clauses(args, this.lastTable(), this.dialect.whereOp);
      this.s.where.push({
        op: 'multi',
        glue: ' OR ',
        clauses: clauses
      });
      return this.pushParams(clauses);
    });
    Select.prototype.orderBy = fluid(function() {
      var args, orderings, _ref;
      args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
      orderings = normalize.orderings(args, this.lastTable(), this.dialect.order);
      return (_ref = this.s.order).push.apply(_ref, orderings);
    });
    Select.prototype.groupBy = function() {
      var field, fields, groupings, _ref;
      fields = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
      groupings = (function() {
        var _i, _len, _results;
        _results = [];
        for (_i = 0, _len = fields.length; _i < _len; _i++) {
          field = fields[_i];
          _results.push(normalize.fieldAndTable({
            table: this.lastTable(),
            field: field
          }));
        }
        return _results;
      }).call(this);
      return (_ref = this.s.groupings).push.apply(_ref, groupings);
    };
    Select.prototype.limit = fluid(function(l) {
      return this.s.limit = l;
    });
    return Select;
  })();
  exports.normalize = normalize = {
    clauses: function(clauses, table, normalizeOp) {
      var clause, constraint, fld, normalized, op, val, _i, _len;
      normalized = [];
      for (_i = 0, _len = clauses.length; _i < _len; _i++) {
        clause = clauses[_i];
        for (fld in clause) {
          constraint = clause[fld];
          if ('object' === typeof constraint) {
            for (op in constraint) {
              val = constraint[op];
              op = normalizeOp(op);
              normalized.push(normalize.fieldAndTable({
                field: fld,
                op: op,
                value: val,
                table: table
              }));
            }
          } else {
            normalized.push(normalize.fieldAndTable({
              field: fld,
              op: '=',
              value: constraint,
              table: table
            }));
          }
        }
      }
      return normalized;
    },
    orderings: function(orderings, table) {
      var add, direction, field, normalized, ordering, _i, _len, _ref;
      normalized = [];
      add = function(field, direction) {
        direction = (function() {
          switch ((direction || '').toLowerCase()) {
            case 'asc':
            case 'ascending':
              return 'ASC';
            case 'desc':
            case 'descending':
              return 'DESC';
            case '':
              return '';
            default:
              throw new Error("Unsupported ordering direction " + direction);
          }
        })();
        return normalized.push(normalize.fieldAndTable({
          field: field,
          table: table,
          direction: direction
        }));
      };
      for (_i = 0, _len = orderings.length; _i < _len; _i++) {
        ordering = orderings[_i];
        if ('string' === typeof ordering) {
          _ref = ordering.split(/\ +/), field = _ref[0], direction = _ref[1];
          add(field, direction);
        } else {
          for (field in ordering) {
            direction = ordering[field];
            add(field, direction);
          }
        }
      }
      return normalized;
    },
    fieldAndTable: function(normalized) {
      var field, table, _ref;
      _ref = normalized.field.split('.'), table = _ref[0], field = _ref[1];
      if (field != null) {
        normalized.table = table;
        normalized.field = field;
      }
      return normalized;
    }
  };
  exports.from = function(tbl, fields, opts) {
    var select;
    if (tbl.constructor === Select) {
      throw new Error("Inner queries not supported yet");
    }
    if (!(opts != null) && (fields != null) && fields.constructor !== Array) {
      opts = fields;
      fields = null;
    }
    select = new Select(tbl, opts);
    if (fields != null) {
      select.fields.apply(select, fields);
    }
    return select;
  };
}).call(this);
