(function() {
  var AND, BOUND_PARAM, DEFAULT_JOIN, JOIN_TYPES, LITERAL, OR, QUOTED, fieldList, group, joins, limit, order, renderBoundParam, renderClause, sys, where;
  var __indexOf = Array.prototype.indexOf || function(item) {
    for (var i = 0, l = this.length; i < l; i++) {
      if (this[i] === item) return i;
    }
    return -1;
  };
  sys = require('sys');
  AND = ' AND ';
  OR = ' OR ';
  JOIN_TYPES = ['INNER', 'CROSS', 'FULL OUTER', 'LEFT', 'LEFT INNER', 'LEFT OUTER', 'RIGHT', 'RIGHT INNER', 'RIGHT OUTER'];
  QUOTED = 'quoted';
  LITERAL = 'literal';
  BOUND_PARAM = '?';
  DEFAULT_JOIN = "INNER";
  exports.renderSelect = function(qs) {
    var ret;
    ret = "SELECT " + (fieldList(qs)) + " FROM ";
    ret += joins(qs);
    ret += where(qs);
    ret += group(qs);
    ret += order(qs);
    return ret += limit(qs);
  };
  fieldList = function(qs) {
    var fields, tbl, tbl_fields, _ref;
    fields = [];
    _ref = qs.fields;
    for (tbl in _ref) {
      tbl_fields = _ref[tbl];
      if (tbl_fields.length) {
        tbl_fields.forEach(function(f) {
          return fields.push("" + tbl + "." + f);
        });
      } else {
        fields.push("" + tbl + ".*");
      }
    }
    return fields.join(', ');
  };
  joins = exports.joins = function(qs) {
    var alias, clause, i, ret, table, tables, type;
    i = 0;
    tables = (function() {
      var _i, _len, _ref, _ref2, _results;
      _ref = qs.tableStack;
      _results = [];
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        _ref2 = _ref[_i], table = _ref2[0], alias = _ref2[1], type = _ref2[2], clause = _ref2[3];
        if (type === 'NOP') {
          continue;
        }
        ret = i++ ? "" + (type.toUpperCase()) + " JOIN " + table : table;
        if (table !== alias) {
          ret += " AS " + alias;
        }
        if (clause != null) {
          ret += " ON " + (renderClause(clause, function(v) {
            return v;
          }));
        }
        _results.push(ret);
      }
      return _results;
    })();
    return tables.join(' ');
  };
  where = exports.where = function(qs) {
    if (qs.where.length) {
      return " WHERE " + (renderClause(qs.where));
    } else {
      return "";
    }
  };
  group = exports.group = function(qs) {
    if (qs.groupings.length) {
      return " GROUP BY " + (qs.groupings.map(function(g) {
        return g.table + '.' + g.field;
      }).join(', '));
    } else {
      return "";
    }
  };
  order = exports.order = function(qs) {
    if (qs.order.length) {
      return ' ORDER BY ' + qs.order.map(function(o) {
        return o.table + '.' + o.field + (o.direction ? ' ' + o.direction : '');
      }).join(', ');
    } else {
      return "";
    }
  };
  limit = function(qs) {
    if (qs.limit != null) {
      return " LIMIT " + qs.limit;
    } else {
      return "";
    }
  };
  renderBoundParam = function(v) {
    if (v && v.constructor === Array) {
      return "(" + (v.map(function() {
        return BOUND_PARAM;
      }).join(', ')) + ")";
    } else {
      return BOUND_PARAM;
    }
  };
  exports.renderClause = renderClause = function(input, renderValue) {
    var render;
    if (renderValue == null) {
      renderValue = renderBoundParam;
    }
    render = function(clause) {
      if ((clause != null) && clause.constructor === Array) {
        return "" + (clause.map(render).join(' AND '));
      } else if (typeof clause === 'object') {
        if (clause.op === 'multi') {
          return "(" + (clause.clauses.map(render).join(clause.glue)) + ")";
        } else {
          return "" + clause.table + "." + clause.field + " " + clause.op + " " + (renderValue(clause.value));
        }
      } else {
        throw new Error("Unexpected clause type, this is probably a bug");
      }
    };
    return render(input);
  };
  exports.joinOp = exports.whereOp = function(op) {
    switch (op.toLowerCase()) {
      case 'ne':
      case '!=':
      case '<>':
        return '!=';
      case 'eq':
      case '=':
        return '=';
      case 'lt':
      case '<':
        return '<';
      case 'gt':
      case '>':
        return '>';
      case 'lte':
      case '<=':
        return '<=';
      case 'gte':
      case '>=':
        return '>=';
      case 'in':
        return 'IN';
      default:
        throw new Error("Unsupported comparison operator: " + op);
    }
  };
  exports.joinType = function(type) {
    if (type == null) {
      return DEFAULT_JOIN;
    }
    type = type.toUpperCase();
    if (__indexOf.call(JOIN_TYPES, type) >= 0) {
      return type;
    }
    throw new Error("Unsupported join type: " + type);
  };
}).call(this);
